import bpy
import math
import random


#Constants
dimension      = {'X': 40,  #Length (mm)
                  'Y': 10,  #Width (mm)
                  'Z': 5}   #Height (mm)
gradientLength = 35         #Gradient length (mm)
dotSize        = {'X': 1,   #Nozzle size (mm)
                  'Y': 1,   #Nozzle size (mm)
                  'Z': 0.8} #Layer height (mm)


#Methods

#Generate vertices
def getVerts(dimensionNumber, dotSize):
    verts = []
    for x in range(0, dimensionNumber['X'] + 1):
        for y in range(0, dimensionNumber['Y'] + 1):
            for z in range(0, dimensionNumber['Z'] + 1):
                verts.append( (x * dotSize['X'], y * dotSize['Y'], z * dotSize['Z']) )
    return verts
    
#Generate gradient points, assuming x as the gradient direction
def getGradient(dimensionNumber, gradientLength):
    gradientNumber = math.floor(gradientLength / dotSize['X'])
    xMin = (dimensionNumber['X'] - gradientNumber) / 2
    xMax = (dimensionNumber['X'] + gradientNumber) / 2 
    gradient = []
    for x in range(0, dimensionNumber['X']):
        xArray = []
        if (x > xMax):
            alpha = 1
        elif(x < xMin):
            alpha = 0
        else:
            alpha = repartitionFunction(x, xMin, 0, xMax, 1)
        for y in range(0, dimensionNumber['Y']):
            yArray = []
            for z in range(0, dimensionNumber['Z']):
                if (random.random() < alpha):
                    point = True
                else:
                    point = False
                yArray.append(point)
            xArray.append(yArray)
        gradient.append(xArray)
    
    return gradient

#Generate faces
def getFaces(dimensionNumber, gradient):
    faces = []
    xMax = dimensionNumber['X']
    yMax = dimensionNumber['Y']
    zMax = dimensionNumber['Z']
    for x in range(0, xMax):
        for y in range(0, yMax):
            for z in range(0, zMax):
                if (gradient[x][y][z]):
                    n = z + y * (zMax + 1) + x * (yMax + 1) * (zMax + 1)
                    cube = [
                        n,
                        n + (zMax + 1) * (yMax + 1),
                        n + (zMax + 1) * (yMax + 2),
                        n + zMax + 1,
                        n + 1,
                        n + 1 + (zMax + 1) * (yMax + 1),
                        n + 1 + (zMax + 1) * (yMax + 2),
                        n + zMax + 2
                    ]
                    if (x == 0 or not gradient[x - 1][y][z]):
                        faces.append( (cube[0], cube[4], cube[7], cube[3]) )
                    if (y == 0 or not gradient[x][y - 1][z]):
                        faces.append( (cube[0], cube[1], cube[5], cube[4]) )
                    if (z == 0 or not gradient[x][y][z - 1]):
                        faces.append( (cube[0], cube[1], cube[2], cube[3]) )
                    if (x == dimensionNumber['X'] - 1 or not gradient[x + 1][y][z]):
                        faces.append( (cube[1], cube[2], cube[6], cube[5]) )
                    if (y == dimensionNumber['Y'] - 1 or not gradient[x][y + 1][z]):
                        faces.append( (cube[2], cube[3], cube[7], cube[6]) )
                    if (z == dimensionNumber['Z'] - 1 or not gradient[x][y][z + 1]):
                        faces.append( (cube[4], cube[5], cube[6], cube[7]) )
    return faces
    
#Make gradient dot object
def makeGradient(dimension, dotSize, gradientLength, gradient, name):
    dimensionNumber = {}
    for key, value in dimension.items():
        dimensionNumber[key] = math.floor(value / dotSize[key])
    verts = getVerts(dimensionNumber, dotSize)
    faces = getFaces(dimensionNumber, gradient)
    
    #Create mesh and object
    mesh = bpy.data.meshes.new(name)
    object = bpy.data.objects.new(name, mesh)
    
    #Set mesh location
    bpy.context.scene.cursor_location = ( 
        - dimensionNumber['X'] * dotSize['X'] / 2, 
        - dimensionNumber['Y'] * dotSize['Y'] / 2, 
        0)
    object.location = bpy.context.scene.cursor_location
    bpy.context.scene.objects.link(object)
    
    #Create mesh from data
    mesh.from_pydata(verts, [], faces)
    mesh.update(calc_edges = True)

#Make gradient and its boolean negation
def makeDualGradient(dimension, dotSize, gradientLength):
    dimensionNumber = {}
    for key, value in dimension.items():
        dimensionNumber[key] = math.floor(value / dotSize[key])
    gradient = getGradient(dimensionNumber, gradientLength)
    makeGradient(dimension, dotSize, gradientLength, gradient, 'Extruder 0')
    for x in range(0, dimensionNumber['X']):
        for y in range(0, dimensionNumber['Y']):
            for z in range(0, dimensionNumber['Z']):
                gradient[x][y][z] = not gradient[x][y][z]
    makeGradient(dimension, dotSize, gradientLength, gradient, 'Extruder 1')

#Delete all objects from the scene
def deleteAll():
    selectAll() 
    bpy.ops.object.delete()
    for item in bpy.data.meshes:
      bpy.data.meshes.remove(item)
    return True

#Select all objects in the scene
def selectAll():
    bpy.ops.object.select_all(action='SELECT')
    return True 
    
#Linear repartition function
def repartitionFunction(x, x0, y0, x1, y1):
    return (y1 - y0) / (x1 - x0) * (x - x0) + y0
    
    
#Script    
deleteAll()
makeDualGradient(dimension, dotSize, gradientLength)
